require 'frontend_compat'

class WebuiProject < ActiveXML::Node

  def self.transport
    ActiveXML::api
  end

  default_find_parameter :name

  attr_accessor :bf_updated
  attr_accessor :pf_updated
  attr_accessor :df_updated
  attr_accessor :uf_updated

  #cache variables
  attr_accessor :my_repositories, :my_repo_hash
  attr_accessor :my_architectures

  handles_xml_element 'project'

  def self.make_stub(opt)
    doc = ActiveXML::Node.new('<project/>')
    doc.set_attribute('name', opt[:name])
    doc.add_element 'title'
    doc.add_element 'description'
    doc
  end

  #check if named project exists
  def self.exists?(pro_name)
    Project.where(name: pro_name).exists?
  end

  attr_writer :api_obj
  def api_obj
    @api_obj ||= Project.find_by_name(to_s)
  end

  #check if named project comes from a remote OBS instance
  def self.is_remote?(pro_name)
    Project.where(name: pro_name).where('remoteurl is not null').exists?
  end
  
  def to_s
    name
  end

  def to_param
    name
  end

  def title
    value(:title)
  end

  def description
    value(:description)
  end

  def add_path_to_repository(opt={})
    return nil if opt == {}
    repository = self.find_first("//repository[@name='#{opt[:reponame]}']")

    unless opt[:repo_path].blank?
      opt[:repo_path] =~ /(.*)\/(.*)/
      repository.each(:path) do |path| # Check if the path to add is already existant
        return false if path.value(:project) == $1 and path.value(:repository) == $2
      end

      param = self.add_element('path', :project => $1, :repository => $2)
      # put it on top
      first = repository.each.first
      if first != param
        first.move_after(param)
      end
    end
  end

  def add_repository( opt={} )
    return nil if opt == {}
    repository = add_element 'repository', 'name' => opt[:reponame]

    unless opt[:repo_path].blank?
      opt[:repo_path] =~ /(.*)\/(.*)/
      repository.add_element 'path', 'project' => $1, 'repository' => $2
    end

    opt[:arch].to_a.each do |arch_text,dummy|
      arch = repository.add_element 'arch'
      arch.text = arch_text
    end
  end

  def move_path_in_repository(repository_name, path_project, path_repository, direction)
    return nil if not repository_name
    return nil if not path_project
    return nil if not path_repository

    repository = self.find_first("//repository[@name='#{repository_name}']")
    case direction
    when :up
      move_path_up_in_repository(repository, path_project, path_repository)
    when :down
      move_path_down_in_repository(repository, path_project, path_repository)
    end if repository
  end

  def remove_path_from_target( repository, path_project, path_repository )
    return nil if not repository
    return nil if not path_project
    return nil if not path_repository

    path = self.find_first("//repository[@name='#{repository}']/path[@project='#{path_project}'][@repository='#{path_repository}']")
    delete_element path if path
  end

  def remove_repository( repository )
    return nil if not repository
    return nil if not self.has_element? :repository

    repository = self.find_first("//repository[@name='#{repository}']")
    delete_element repository if repository
  end

  def release_repository( repository, target=nil )
    # target is optional and may come as string "project/targetrepository"

    arguments = {:project => self.name, :cmd => 'release'}
    if target
      a=target.split(/\//)
      arguments[:targetproject] = a[0]
      arguments[:targetrepository] = a[1]
    end

    begin
      fc = FrontendCompat.new
      answer = fc.do_post(nil, arguments)
      doc = ActiveXML::Node.new(answer)
      doc.each('/collection/project') {|e| result << e.value('name')}
    rescue ActiveXML::Transport::NotFoundError
      # No answer is ok, it only means no linking projects...
    end
    return result
  end


  def add_maintained_project(maintained_project)
    return nil if not maintained_project
    add_element('maintenance') if not has_element?('maintenance')
    self.find_first(:maintenance).add_element('maintains', 'project' => maintained_project)
  end

  def remove_maintained_project(maintained_project)
    return nil if not maintained_project
    return nil if not has_element?('maintenance')
    self.find_first(:maintenance).delete_element("maintains[@project='#{maintained_project}']")
  end

  #get all architectures used in this project
  def architectures
    self.my_architectures ||= api_obj.repositories.joins(:architectures).pluck('distinct architectures.name')
  end

  def repositories
    ret = Array.new
    self.each(:repository) {|repo| ret << repo.value(:name) }
    ret
  end

  def repository
    repo_hash = Hash.new
    self.each(:repository) {|repo| repo_hash[repo.value(:name)] = repo}
    repo_hash
  end
    
  def users(role = nil)
    rels = api_obj.relationships
    rels = rels.where(role: Role.rolecache[role]) if role
    users = rels.users.pluck(:user_id)
    User.where(id: users.flatten.uniq)
  end

  def groups(role = nil)
    rels = api_obj.relationships
    rels = rels.where(role: Role.rolecache[role]) if role
    Group.where(id: rels.groups.pluck(:group_id).uniq)
  end

  def name
    @name ||= to_hash['name']
  end

  def name=(s)
    @name = s
  end

  def to_param
    name
  end

  def project_type
    api_obj.project_type
  end

  def set_project_type(project_type)
    if %w(maintenance maintenance_incident standard).include?(project_type)
      set_attribute('kind', project_type)
      return true
    end
    return false
  end

  def is_remote?
    api_obj.remoteurl.present?
  end

  def self.attributes(project_name)
    path = "/source/#{project_name}/_attribute/"
    res = ActiveXML::api.direct_http(URI("#{path}"))
    return Collection.new(res)
  end

  def attributes
    return WebuiProject.attributes(self.name)
  end

  def self.has_attribute?(project_name, attribute_namespace, attribute_name)
    self.attributes(project_name).each do |attr|
      return true if attr.namespace == attribute_namespace && attr.name == attribute_name
    end
    return false
  end

  def issues
    return Rails.cache.fetch("changes_and_patchinfo_issues_#{self.name}2", :expires_in => 5.minutes) do
      issues = WebuiProject.find(:issues, :name => self.name, :expires_in => 5.minutes)
      if issues
        changes_issues, patchinfo_issues = {}, {}
        issues.each(:package) do |package|
          package.each(:issue) do |issue|
            if package.value('name') == 'patchinfo'
              patchinfo_issues[issue.value('label')] = issue
            else
              changes_issues[issue.value('label')] = issue
            end
          end
        end
        missing_issues, optional_issues = {}, {}
        changes_issues.each do |label, issue|
          optional_issues[label] = issue unless patchinfo_issues.has_key?(label)
        end
        patchinfo_issues.each do |label, issue|
          missing_issues[label] = issue unless changes_issues.has_key?(label)
        end
        {:changes => changes_issues, :patchinfo => patchinfo_issues, :missing => missing_issues, :optional => optional_issues}
      else
        {}
      end
    end
  end

  def is_locked?
    api_obj.is_locked?
  end

  def self.find(name, opts = {})
    name = name.to_param
    begin
      ap = Project.get_by_name(name)
    rescue Project::UnknownObjectError
      return nil
    end
    if ap.kind_of? Project
      p = WebuiProject.new '<project/>'
      p.api_obj = ap
      p.name = ap.name
      p.instance_variable_set('@init_options', name: name)
    else
      p = super
    end
    p
  end

  def parse(data)
    if @api_obj
      data = @api_obj.to_axml
    end
    super(data)
  end

  private

  def move_path_up_in_repository(repository, path_project, path_repository)
    previous_path = nil
    repository.each(:path).each do |path|
      if path.value(:project) == path_project and path.value(:repository) == path_repository
        path.move_before(previous_path) if previous_path
        return
      end
      previous_path = path
    end
  end

  def move_path_down_in_repository(repository, path_project, path_repository)
    next_path = nil
    repository.each(:path).reverse.each do |path|
      if path.value(:project) == path_project and path.value(:repository) == path_repository
        path.move_after(next_path) if next_path
        return
      end
      next_path = path
    end
  end
end
